#pragma once

#include <memory>

#include "future.h"

namespace asyncpp {
namespace __future {

    template<
        typename T_future,
        typename T_lambda>
    struct and_then_future final :
        public base_future<
            decltype(std::declval<T_lambda>()(std::declval<typename std::decay_t<T_future>::value_type>())),
            and_then_future<T_future, T_lambda>
        >
    {
    public:
        using lambda_type               = T_lambda;
        using first_future_type         = T_future;
        using first_value_type          = typename std::decay_t<first_future_type>::value_type;
        using second_future_type        = decltype(std::declval<lambda_type>()(std::declval<first_value_type>()));
        using second_future_type_ptr    = std::unique_ptr<second_future_type>;
        using value_type                = typename second_future_type::value_type;
        using this_type                 = and_then_future<value_type, lambda_type>;
        using base_future_type          = base_future<value_type, this_type>;
        using result_type               = typename base_future_type::result_type;

    private:
        lambda_type             _lambda;
        first_future_type       _first;
        second_future_type_ptr  _second;

    public:
        /**
         * @brief Constructor.
         */
        template<
            typename X_future,
            typename X_lambda>
        inline and_then_future(
            X_future&& p_first,
            X_lambda&& p_lambda)
            : _first (std::forward<X_future>(p_first))
            , _lambda(std::forward<X_lambda>(p_lambda))
            { }

        inline and_then_future(and_then_future &&) = default;
        inline and_then_future(and_then_future const &) = default;

    public: /* future */
        /**
         * @brief Poll the result from the future.
         */
        inline result_type poll()
        {
            while (true)
            {
                if (_second)
                {
                    return _second->poll();
                }
                else
                {
                    auto r = _first.poll();
                    if (!r)
                        return result_type::not_ready();

                    _second = std::make_unique<second_future_type>(r.call(_lambda));
                }
            }
        }
    };

    /* and_then_impl */

    template<
        chaining_mode X_mode,
        typename X_self,
        typename X_lambda>
    auto and_then_impl(X_self&& p_self, X_lambda&& p_lambda)
    {
        using lambda_type           = X_lambda;
        using self_type             = X_self;
        using derived_type          = typename std::decay_t<self_type>::derived_type;
        using derived_storage_type  = storage_type_t<X_mode, derived_type>;
        using derived_ref_type      = std::conditional_t<
                                        std::is_const_v<std::remove_reference_t<self_type>>,
                                        derived_type const &,
                                        derived_type &>;
        using derived_forward_type  = forward_type_t<X_mode, derived_ref_type>;
        using and_then_future_type  = __future::and_then_future<derived_storage_type, lambda_type>;

        auto& self = static_cast<derived_ref_type>(p_self);

        return and_then_future_type(
            std::forward<derived_forward_type>(self),
            std::forward<X_lambda>            (p_lambda));
    }

} }

namespace asyncpp
{

    /* future_base::and_then */

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode,
        typename X_lambda>
    auto base_future<T_value, T_derived>
        ::and_then(X_lambda&& p_lambda) &
    {
        return __future::and_then_impl<X_mode>(*this, std::forward<X_lambda>(p_lambda));
    }

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode,
        typename X_lambda>
    auto base_future<T_value, T_derived>
        ::and_then(X_lambda&& p_lambda) &&
    {
        static_assert(
            X_mode != ref,
            "Can not store rvalue reference as lvalue reference!");

        return __future::and_then_impl<X_mode>(std::move(*this), std::forward<X_lambda>(p_lambda));
    }

}
