#pragma once

#include "future.h"

namespace asyncpp {
namespace __future {

    template<
        typename T_future,
        typename T_lambda>
    struct map_future final :
        public base_future<
            typename std::decay_t<T_future>::value_type,
            map_future<T_future, T_lambda>
        >
    {
    public:
        using future_type       = T_future;
        using lambda_type       = T_lambda;
        using future_value_type = typename std::decay_t<future_type>::value_type;
        using value_type        = decltype(std::declval<lambda_type>()(std::declval<future_value_type>()));
        using this_type         = map_future<future_type, lambda_type>;
        using base_future_type  = base_future<value_type, this_type>;
        using result_type       = typename base_future_type::result_type;

    private:
        future_type _future;
        lambda_type _lambda;

    public:
        /**
         * @brief Constructor.
         */
        template<
            typename X_future,
            typename X_lambda>
        inline map_future(
            X_future&& p_future,
            X_lambda&& p_lambda)
            : _future(std::forward<X_future>(p_future))
            , _lambda(std::forward<X_lambda>(p_lambda))
            { }

        inline map_future(map_future &&) = default;
        inline map_future(map_future const &) = default;

    public: /* future */
        /**
         * @brief Poll the result from the future.
         */
        inline result_type poll()
        {
            auto r = _future.poll();
            return r
                ? result_type::ready(r.call(_lambda))
                : result_type::not_ready();
        }
    };

    /* map_impl */

    template<
        chaining_mode X_mode,
        typename X_self,
        typename X_lambda>
    auto map_impl(X_self&& p_self, X_lambda&& p_lambda)
    {
        using lambda_type           = X_lambda;
        using self_type             = X_self;
        using derived_type          = typename std::decay_t<self_type>::derived_type;
        using derived_storage_type  = storage_type_t<X_mode, derived_type>;
        using derived_ref_type      = std::conditional_t<
                                        std::is_const_v<std::remove_reference_t<self_type>>,
                                        derived_type const &,
                                        derived_type &>;
        using derived_forward_type  = forward_type_t<X_mode, derived_ref_type>;
        using map_future_type       = __future::map_future<derived_storage_type, lambda_type>;

        auto& self = static_cast<derived_ref_type>(p_self);

        return map_future_type(
            std::forward<derived_forward_type>(self),
            std::forward<X_lambda>            (p_lambda));
    }

} }

namespace asyncpp
{

    /* future_base::map */

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode,
        typename X_lambda>
    auto base_future<T_value, T_derived>
        ::map(X_lambda&& p_lambda) &
    {
        return __future::map_impl<X_mode>(*this, std::forward<X_lambda>(p_lambda));
    }

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode,
        typename X_lambda>
    auto base_future<T_value, T_derived>
        ::map(X_lambda&& p_lambda) &&
    {
        static_assert(
            X_mode != ref,
            "Can not store rvalue reference as lvalue reference!");

        return __future::map_impl<X_mode>(std::move(*this), std::forward<X_lambda>(p_lambda));
    }

}
