#pragma once

#include "result.h"

namespace asyncpp
{

    namespace __impl
    {

        /* result_ready */

        template<typename T_value>
        template<typename... T_args>
        result_ready<T_value>
            ::result_ready(T_args&&... p_args)
                : value(std::forward<T_args>(p_args)...)
                { }

        /* result */

        template<bool for_stream, typename T_value>
        result<for_stream, T_value>
            ::result(storage_type&& p_storage)
                : _storage(std::move(p_storage))
            { }

        template<bool for_stream, typename T_value>
        auto& result<for_stream, T_value>
            ::not_ready()
        {
            static const result ret(storage_type(not_ready_type { }));
            return ret;
        }

        template<bool for_stream, typename T_value>
        template<typename... X_args>
        auto result<for_stream, T_value>
            ::ready(X_args&&... p_args)
            { return result(storage_type(ready_type(std::forward<X_args>(p_args)...))); }

        template<bool for_stream, typename T_value>
        template<bool X, typename>
        auto& result<for_stream, T_value>
            ::done()
        {
            static const result ret(storage_type(done_type { }));
            return ret;
        }

        template<bool for_stream, typename T_value>
        template<bool X, typename>
        result_status result<for_stream, T_value>
            ::status() const
        {
            if (is_not_ready())
                return result_status::not_ready;
            else if (is_ready())
                return result_status::ready;
            else if (is_done())
                return result_status::done;
            else
                return result_status::unknown;
        }

        template<bool for_stream, typename T_value>
        bool result<for_stream, T_value>
            ::is_not_ready() const
            { return std::holds_alternative<not_ready_type>(_storage); }

        template<bool for_stream, typename T_value>
        bool result<for_stream, T_value>
            ::is_ready() const
            { return std::holds_alternative<ready_type>(_storage); }

        template<bool for_stream, typename T_value>
        template<bool X, typename>
        bool result<for_stream, T_value>
            ::is_done() const
            { return std::holds_alternative<done_type>(_storage); }

        template<bool for_stream, typename T_value>
        typename result<for_stream, T_value>::reference_type
        result<for_stream, T_value>
            ::value()
            { return std::get<ready_type>(_storage).value; }

        template<bool for_stream, typename T_value>
        typename result<for_stream, T_value>::const_reference_type
        result<for_stream, T_value>
            ::value() const
            { return std::get<ready_type>(_storage).value; }

        template<bool for_stream, typename T_value>
        template<typename X_lambda>
        auto result<for_stream, T_value>
            ::call(const X_lambda& p_lambda)
            { return p_lambda(value()); }

        template<bool for_stream, typename T_value>
        template<typename X_lambda>
        auto result<for_stream, T_value>
            ::call(const X_lambda& p_lambda) const
            { return p_lambda(value()); }

        template<bool for_stream, typename T_value>
        result<for_stream, T_value>
            ::operator bool() const
            { return is_ready(); }

        template<bool for_stream, typename T_value>
        typename result<for_stream, T_value>::pointer_type
        result<for_stream, T_value>
            ::operator-> ()
            { return &value(); }

        template<bool for_stream, typename T_value>
        typename result<for_stream, T_value>::reference_type
        result<for_stream, T_value>
            ::operator* ()
            { return value(); }

        template<bool for_stream, typename T_value>
        typename result<for_stream, T_value>::const_pointer_type
        result<for_stream, T_value>
            ::operator-> () const
            { return &value(); }

        template<bool for_stream, typename T_value>
        typename result<for_stream, T_value>::const_reference_type
        result<for_stream, T_value>
            ::operator* () const
            { return value(); }

        /* result<void> */

        template<bool for_stream>
        struct result<for_stream, void>
        {
        public:
            using value_type = void;

        private:
            result_status _status;

        private:
            inline result(result_status p_status)
                : _status(p_status)
                { }

        public:
            static inline auto& not_ready()
            {
                static const result ret { result_status::not_ready };
                return ret;
            }

            template<typename... X_args>
            static inline auto& ready(X_args&&... p_args)
            {
                static const result ret { result_status::ready };
                return ret;
            }

            template<
                bool X = for_stream,
                typename = std::enable_if_t<X>>
            static inline auto& done()
            {
                static const result ret { result_status::done };
                return ret;
            }

        public:
            template<
                bool X = for_stream,
                typename = std::enable_if_t<X>>
            inline result_status status() const
                { return _status; }

            inline bool is_not_ready() const
                { return _status == result_status::not_ready; }

            inline bool is_ready() const
                { return _status == result_status::ready; }

            template<
                bool X = for_stream,
                typename = std::enable_if_t<X>>
            inline bool is_done() const
                { return _status == result_status::done; }

            inline void value()
                { throw std::runtime_error("'void' result does not store any value!"); }

            template<typename X_lambda>
            inline auto call(const X_lambda& p_lambda)
                { return p_lambda(); }

            inline operator bool() const
                { return _status == result_status::ready; }

            inline void operator* ()
                { value(); }
        };

    }

}
