#pragma once

#include <memory>

#include "stream.h"

namespace asyncpp {
namespace __stream {

    template<
        typename T_stream>
    struct flatten_stream final :
        public base_stream<
            typename T_stream::value_type::value_type,
            flatten_stream<T_stream>
        >
    {
    public:
        using stream_type           = T_stream;
        using inner_stream_type     = typename stream_type::value_type;
        using inner_stream_ptr_u    = std::unique_ptr<inner_stream_type>;
        using value_type            = typename inner_stream_type::value_type;
        using this_type             = flatten_stream<stream_type>;
        using base_stream_type      = base_stream<value_type, this_type>;
        using result_type           = typename base_stream_type::result_type;

    private:
        stream_type         _stream;
        inner_stream_ptr_u  _inner;

    public:
        /**
         * @brief Constructor.
         */
        template<
            typename X_stream>
        inline flatten_stream(
            X_stream&& p_stream)
            : _stream   (std::forward<X_stream>(p_stream))
            , _inner    ()
            { }

        inline flatten_stream(flatten_stream &&) = default;
        inline flatten_stream(flatten_stream const &) = default;

    public: /* future */
        /**
         * @brief Poll the result from the future.
         */
        inline result_type poll()
        {
            while (true)
            {
                if (_inner)
                {
                    auto r = _inner->poll();

                    if (!r.is_done())
                        return r;

                    _inner.reset();
                }

                auto r = _stream.poll();
                if (r.is_done())
                    return result_type::done();

                if (r.is_not_ready())
                    return result_type::not_ready();

                _inner = std::make_unique<inner_stream_type>(std::move(*r));
            }
        }
    };

    /* flatten_impl */

    template<
        chaining_mode X_mode,
        typename X_self>
    auto map_impl(X_self&& p_self)
    {
        using self_type             = X_self;
        using derived_type          = typename std::decay_t<self_type>::derived_type;
        using derived_storage_type  = storage_type_t<X_mode, derived_type>;
        using derived_ref_type      = std::conditional_t<
                                        std::is_const_v<std::remove_reference_t<self_type>>,
                                        derived_type const &,
                                        derived_type &>;
        using derived_forward_type  = forward_type_t<X_mode, derived_ref_type>;
        using flatten_stream_type   = __stream::flatten_stream<derived_storage_type>;

        auto& self = static_cast<derived_ref_type>(p_self);

        return flatten_stream_type(
            std::forward<derived_forward_type>(self));
    }

} }

namespace asyncpp
{

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode>
    auto base_stream<T_value, T_derived>
        ::flatten() &
    {
        return __stream::map_impl<X_mode>(*this);
    }

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode>
    auto base_stream<T_value, T_derived>
        ::flatten() &&
    {
        static_assert(
            X_mode != ref,
            "Can not store rvalue reference as lvalue reference!");

        return __stream::map_impl<X_mode>(std::move(*this));
    }

}
