#pragma once

#include "stream.h"

namespace asyncpp {
namespace __stream {

    template<
        typename T_stream,
        typename T_lambda>
    struct for_each_future final :
        public base_future<
            void,
            for_each_future<T_stream, T_lambda>
        >
    {
    public:
        using stream_type       = T_stream;
        using lambda_type       = T_lambda;
        using value_type        = void;
        using this_type         = for_each_future<stream_type, lambda_type>;
        using base_future_type  = base_future<value_type, this_type>;
        using result_type       = typename base_future_type::result_type;

    private:
        stream_type _stream;
        lambda_type _lambda;

    public:
        /**
         * @brief Constructor.
         */
        template<
            typename X_stream,
            typename X_lambda>
        inline for_each_future(
            X_stream&& p_stream,
            X_lambda&& p_lambda)
            : _stream(std::forward<X_stream>(p_stream))
            , _lambda(std::forward<X_lambda>(p_lambda))
            { }

        inline for_each_future(for_each_future &&) = default;
        inline for_each_future(for_each_future const &) = default;

    public: /* future */
        /**
         * @brief Poll the result from the future.
         */
        inline result_type poll()
        {
            while (true)
            {
                auto r = _stream.poll();
                if (r.is_done())
                    return result_type::ready();

                if (r.is_not_ready())
                    return result_type::not_ready();

                r.call(_lambda);
            }
        }
    };

    template<
        chaining_mode X_mode,
        typename X_self,
        typename X_lambda>
    auto for_each_impl(X_self&& p_self, X_lambda&& p_lambda)
    {
        using self_type             = X_self;
        using lambda_type           = X_lambda;
        using derived_type          = typename std::decay_t<self_type>::derived_type;
        using derived_storage_type  = storage_type_t<X_mode, derived_type>;
        using derived_ref_type      = std::conditional_t<
                                        std::is_const_v<std::remove_reference_t<self_type>>,
                                        derived_type const &,
                                        derived_type &>;
        using derived_forward_type  = forward_type_t<X_mode, derived_ref_type>;
        using for_each_future_type  = __stream::for_each_future<derived_storage_type, lambda_type>;

        auto& self = static_cast<derived_ref_type>(p_self);

        return for_each_future_type(
            std::forward<derived_forward_type>(self),
            std::forward<X_lambda>(p_lambda));
    }

} }

namespace asyncpp
{

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode,
        typename X_lambda>
    auto base_stream<T_value, T_derived>
        ::for_each(X_lambda&& p_lambda) &
    {
        return __stream::for_each_impl<X_mode>(*this, std::forward<X_lambda>(p_lambda));
    }

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode,
        typename X_lambda>
    auto base_stream<T_value, T_derived>
        ::for_each(X_lambda&& p_lambda) &&
    {
        static_assert(
            X_mode != ref,
            "Can not store rvalue reference as lvalue reference!");

        return __stream::for_each_impl<X_mode>(std::move(*this), std::forward<X_lambda>(p_lambda));
    }

}
