#pragma once

#include "stream.h"

namespace asyncpp {
namespace __stream {

    template<
        typename T_stream,
        typename T_lambda>
    struct map_stream final :
        public base_stream<
            decltype(std::declval<T_lambda>()(std::declval<typename T_stream::value_type>())),
            map_stream<T_stream, T_lambda>
        >
    {
    public:
        using stream_type       = T_stream;
        using lambda_type       = T_lambda;
        using inner_value_type  = typename stream_type::value_type;
        using value_type        = decltype(std::declval<lambda_type>()(std::declval<inner_value_type>()));
        using this_type         = map_stream<stream_type, lambda_type>;
        using base_stream_type  = base_stream<value_type, this_type>;
        using result_type       = typename base_stream_type::result_type;

    private:
        stream_type _stream;
        lambda_type _lambda;

    public:
        /**
         * @brief Constructor.
         */
        template<
            typename X_stream,
            typename X_lambda>
        inline map_stream(
            X_stream&& p_stream,
            X_lambda&& p_lambda)
            : _stream(std::forward<X_stream>(p_stream))
            , _lambda(std::forward<X_lambda>(p_lambda))
            { }

        inline map_stream(map_stream &&) = default;
        inline map_stream(map_stream const &) = default;

    public: /* future */
        /**
         * @brief Poll the result from the future.
         */
        inline result_type poll()
        {
            auto r = _stream.poll();
            if (r.is_done())
                return result_type::done();

            if (r.is_ready())
                return result_type::ready(r.call(_lambda));

            return result_type::not_ready();
        }
    };

    template<
        chaining_mode X_mode,
        typename X_self,
        typename X_lambda>
    auto map_impl(X_self&& p_self, X_lambda&& p_lambda)
    {
        using self_type             = X_self;
        using lambda_type           = X_lambda;
        using derived_type          = typename std::decay_t<self_type>::derived_type;
        using derived_storage_type  = storage_type_t<X_mode, derived_type>;
        using derived_ref_type      = std::conditional_t<
                                        std::is_const_v<std::remove_reference_t<self_type>>,
                                        derived_type const &,
                                        derived_type &>;
        using derived_forward_type  = forward_type_t<X_mode, derived_ref_type>;
        using map_stream_type       = __stream::map_stream<derived_storage_type, lambda_type>;

        auto& self = static_cast<derived_ref_type>(p_self);

        return map_stream_type(
            std::forward<derived_forward_type>(self),
            std::forward<X_lambda>(p_lambda));
    }

} }

namespace asyncpp
{

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode,
        typename X_lambda>
    auto base_stream<T_value, T_derived>
        ::map(X_lambda&& p_lambda) &
    {
        return __stream::map_impl<X_mode>(*this, std::forward<X_lambda>(p_lambda));
    }

    template<
        typename T_value,
        typename T_derived>
    template<
        chaining_mode X_mode,
        typename X_lambda>
    auto base_stream<T_value, T_derived>
        ::map(X_lambda&& p_lambda) &&
    {
        static_assert(
            X_mode != ref,
            "Can not store rvalue reference as lvalue reference!");

        return __stream::map_impl<X_mode>(std::move(*this), std::forward<X_lambda>(p_lambda));
    }

}
