#include <gtest/gtest.h>

#include <asyncpp.h>

using namespace ::testing;
using namespace ::asyncpp;

struct delay
    : public base_stream<int, delay>
{
private:
    int const _delay     { 5 };
    int const _threshold { 2 };
    int       _count     { 0 };

public:
    inline delay(int p_delay, int p_threshold, int p_count)
        : _delay    (p_delay)
        , _threshold(p_threshold)
        , _count    (p_count)
        { }

    auto poll()
    {
        if (_count >= _delay)
            return result_type::done();

        ++_count;

        return _count <= _threshold
            ? result_type::not_ready()
            : result_type::ready(_count);
    }
};

TEST(stream_tests, poll)
{
    delay s { 5, 2, 0 };

    auto r0 = s.poll();
    ASSERT_FALSE(r0);

    auto r1 = s.poll();
    ASSERT_FALSE(r1);

    auto r2 = s.poll();
    ASSERT_TRUE (r2);
    ASSERT_EQ   (3, *r2);

    auto r3 = s.poll();
    ASSERT_TRUE (r3);
    ASSERT_EQ   (4, *r3);

    auto r4 = s.poll();
    ASSERT_TRUE (r4);
    ASSERT_EQ   (5, *r4);

    auto r = s.poll();
    ASSERT_FALSE(r);
    ASSERT_TRUE (r.is_done());
}

TEST(stream_tests, for_each)
{
    int i = 0;
    delay d { 5, 0, 0 };
    auto f = d
        .for_each([&i](int x) {
            ++i;
            EXPECT_EQ(i, x);
        });

    auto r = f.poll();
    ASSERT_TRUE(r);
}

TEST(stream_tests, map)
{
    delay d { 5, 0, 0 };
    auto s = d
        .map([](int i){
            return std::to_string(i);
        });

    auto r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, "1");

    r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, "2");

    r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, "3");

    r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, "4");

    r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, "5");

    r = s.poll();
    ASSERT_TRUE(r.is_done());
}

TEST(stream_tests, flatten)
{
    delay d { 3, 0, 0 };
    auto s = d
        .map([](int i){
            return delay { i * 100 + 2, 0, i * 100 };
        })
        .flatten();

    auto r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, 101);

    r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, 102);

    r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, 201);

    r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, 202);

    r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, 301);

    r = s.poll();
    ASSERT_TRUE(r);
    ASSERT_EQ  (*r, 302);

    r = s.poll();
    ASSERT_TRUE(r.is_done());
}
