#include <gtest/gtest.h>

#include "../../helper/now_mock.h"
#include "../../helper/runtime_mock.h"

#include <asyncpp.h>
#include <asyncpp/executor/current_thread.h>

using namespace ::testing;
using namespace ::asyncpp;

struct test
    : public base_future<void, test>
{
public:
    using value_type        = void;
    using this_type         = test;
    using base_future_type  = base_future<void, this_type>;
    using result_type       = typename base_future_type::result_type;

public:
    bool            done    { false };
    asyncpp::task   task;

public:
    inline test() = default;

public:
    inline auto poll()
    {
        task = task::current();
        return done
            ? result_type::ready()
            : result_type::not_ready();
    }
};

struct mock
{
    MOCK_METHOD0(call, void());
};

TEST(current_thread_tests, run)
{
    Sequence s;
    StrictMock<runtime_mock> m;
    executor::current_thread<StrictMock<runtime_mock>&> e(m);

    EXPECT_CALL(m, init_thread());

    test t;

    EXPECT_CALL(m, idle(nullptr))
        .InSequence(s)
        .WillOnce(Invoke([&t](auto){
            t.task.notify();
        }))
        .WillOnce(Invoke([&t](auto){
            t.done = true;
            t.task.notify();
        }));;

    e.run(t);
}

TEST(current_thread_tests, interval)
{
    using mock_type         = StrictMock<mock>;
    using now_mock_type     = StrictMock<now_mock>;
    using runtime_mock_type = StrictMock<runtime_mock>;
    using runtime_type      = timing::timer<runtime_mock_type&>;
    using executor_type     = executor::current_thread<runtime_type&>;

    InSequence          s;
    mock_type           m;
    now_mock_type       nm;
    runtime_mock_type   rm;
    runtime_type        r(rm);
    executor_type       e(r);

    EXPECT_CALL(rm, init_thread());
    EXPECT_CALL(nm, now()) // timing::delay::poll
        .WillOnce(Return(time_point(std::chrono::seconds(0))));
    EXPECT_CALL(m, call());

    EXPECT_CALL(nm, now()) // timing::delay::poll
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    EXPECT_CALL(nm, now()) // timer::idle
        .WillOnce(Return(time_point(std::chrono::seconds(0))));
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(10)))));

    EXPECT_CALL(nm, now()) // timer::idle
        .WillOnce(Return(time_point(std::chrono::seconds(10))));
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(10)))));

    EXPECT_CALL(nm, now()) // timing::delay::poll
        .WillOnce(Return(time_point(std::chrono::seconds(11))));
    EXPECT_CALL(m, call());

    EXPECT_CALL(nm, now()) // timing::delay::poll
        .WillOnce(Return(time_point(std::chrono::seconds(11))));

    EXPECT_CALL(nm, now()) // timer::idle
        .WillOnce(Return(time_point(std::chrono::seconds(15))));
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(20)))));

    EXPECT_CALL(nm, now()) // timer::idle
        .WillOnce(Return(time_point(std::chrono::seconds(49))));
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(20)))));

    EXPECT_CALL(nm, now()) // timing::delay::poll
        .WillOnce(Return(time_point(std::chrono::seconds(49))));
    EXPECT_CALL(m, call());

    EXPECT_CALL(nm, now()) // timing::delay::poll
        .WillOnce(Return(time_point(std::chrono::seconds(49))));
    EXPECT_CALL(m, call());

    EXPECT_CALL(nm, now()) // timing::delay::poll
        .WillOnce(Return(time_point(std::chrono::seconds(49))));
    EXPECT_CALL(m, call());
    EXPECT_CALL(nm, now()) // timing::interval::poll
        .WillOnce(Return(time_point(std::chrono::seconds(49))));

    EXPECT_CALL(nm, now()) // timing::delay::poll
        .WillOnce(Return(time_point(std::chrono::seconds(49))));

    EXPECT_CALL(nm, now()) // timer::idle
        .WillOnce(Return(time_point(std::chrono::seconds(49))));
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(50)))));

    EXPECT_CALL(nm, now()) // timer::idle
        .WillOnce(Return(time_point(std::chrono::seconds(50))));
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(50)))));

    EXPECT_CALL(nm, now()) // timing::interval::poll
        .WillOnce(Return(time_point(std::chrono::seconds(50))));

    e.run(
        timing::interval(
            time_point(),
            std::chrono::seconds(10),
            time_point() + std::chrono::seconds(50))
        .for_each([&m]{
            m.call();
        }));
}
