#include <gtest/gtest.h>

#include <asyncpp.h>
#include <asyncpp/fs.h>

using namespace ::testing;
using namespace ::asyncpp;

struct auto_remove_dir
    : public cppfs::directory
{
public:
    using cppfs::directory::directory;

    inline ~auto_remove_dir()
    {
        if (exists())
            remove(true);
    }
};

class directory_tests
    : public ::testing::Test
{
protected:
    static const std::string& path_test_dir()
    {
        static const std::string value("/tmp/asyncpp/test");
        return value;
    }

    static const std::string& path_test_dir_single()
    {
        static const std::string value("/tmp/asyncpp/test/single");
        return value;
    }

    static const std::string& path_test_dir_parent()
    {
        static const std::string value("/tmp/asyncpp/test/parent");
        return value;
    }

    static const std::string& path_test_dir_child()
    {
        static const std::string value("/tmp/asyncpp/test/parent/child");
        return value;
    }

    void SetUp() override
    {
        cppfs::directory d(path_test_dir());
        if (!d.exists())
            d.create(true);
    }

    void TearDown() override
    {
        cppfs::directory d(path_test_dir());
        if (d.exists())
            d.remove(true);
    }
};

TEST_F(directory_tests, read_dir)
{
    using namespace cppfs;
    using namespace std::string_literals;

    auto s = fs::directory("."s).read();
    auto v = std::vector<std::string>();
    auto r = decltype(s)::result_type::not_ready();

    while ((r = s.poll()).is_ready())
        v.emplace_back(r->path.parts().back());

    ASSERT_TRUE(r.is_done());

    std::sort(v.begin(), v.end(), cppcore::op_invariant_string_less { });
    EXPECT_EQ(v, std::vector<std::string>({
        "asyncpp",
        "CMakeLists.txt",
        "helper",
    }));
}

TEST_F(directory_tests, read_dir_all)
{
    using namespace cppfs;
    using namespace std::string_literals;

    auto s = fs::directory("."s).read_all();
    auto r = decltype(s)::result_type::not_ready();

    auto entries = std::vector<cppfs::directory::entry>();
    while ((r = s.poll()).is_ready())
        entries.emplace_back(*r);

    ASSERT_TRUE(r.is_done());

    std::sort(entries.begin(), entries.end());

    std::vector<std::string> paths;
    paths.reserve(entries.size());
    std::transform(
        entries.begin(),
        entries.end(),
        std::back_inserter(paths),
        [](auto& e) {
            auto s = std::to_string(static_cast<int>(e.type)) + e.path.str();
            return s;
        });

    EXPECT_EQ(paths, std::vector<std::string>({
        "1./asyncpp",
        "1./asyncpp/core",
        "2./asyncpp/core/future_tests.cpp",
        "2./asyncpp/core/result_tests.cpp",
        "2./asyncpp/core/stream_tests.cpp",
        "2./asyncpp/core/task_tests.cpp",
        "1./asyncpp/executor",
        "2./asyncpp/executor/current_thread_tests.cpp",
        "1./asyncpp/fs",
        "2./asyncpp/fs/directory_tests.cpp",
        "1./asyncpp/timing",
        "2./asyncpp/timing/delay_tests.cpp",
        "2./asyncpp/timing/interval_tests.cpp",
        "2./asyncpp/timing/timeout_tests.cpp",
        "2./asyncpp/timing/timer_tests.cpp",
        "1./helper",
        "2./helper/now_mock.h",
        "2./helper/runtime_mock.h",
        "2./helper/test_delay.h",
        "2./CMakeLists.txt",
    }));
}

TEST_F(directory_tests, create_dir)
{
    auto f = fs::directory(path_test_dir_single())
        .create();

    auto_remove_dir d(path_test_dir_single());

    ASSERT_FALSE(d.exists());

    auto r = f.poll();

    ASSERT_TRUE (r);
    ASSERT_TRUE (d.exists());
}

TEST_F(directory_tests, create_dir_all)
{
    auto f = fs::directory(path_test_dir_child())
        .create_all();

    auto_remove_dir d(path_test_dir_child());

    ASSERT_FALSE(d.exists());

    auto r = f.poll();

    ASSERT_TRUE (r);
    ASSERT_TRUE (d.exists());
}

TEST_F(directory_tests, remove_dir)
{
    auto f = fs::directory(path_test_dir_single())
        .remove();

    auto_remove_dir d(path_test_dir_single());
    d.create(true);

    ASSERT_TRUE (d.exists());

    auto r = f.poll();

    ASSERT_TRUE (r);
    ASSERT_FALSE(d.exists());
}

TEST_F(directory_tests, remove_dir_all)
{
    auto f = fs::directory(path_test_dir_child())
        .remove_all();

    auto_remove_dir d(path_test_dir_child());
    d.create(true);

    ASSERT_TRUE (d.exists());

    auto r = f.poll();

    ASSERT_TRUE (r);
    ASSERT_FALSE(d.exists());
}
