#include <gtest/gtest.h>

#include "../../helper/now_mock.h"

#include <asyncpp.h>

using namespace ::testing;
using namespace ::asyncpp;
using namespace ::asyncpp::timing;

struct test_future
    : public base_future<int, test_future>
{
public:
    using value_type        = int;
    using this_type         = test_future;
    using base_future_type  = base_future<value_type, this_type>;

public:
    int i;

    inline test_future(int p_i)
        : i(p_i)
        { }

public:
    inline result_type poll()
    {
        return i
            ? result_type::ready(i)
            : result_type::not_ready();
    }
};

auto make_test_future(int i)
{
    test_future ret(i);
    return ret;
}

struct test_stream
    : public base_stream<int, test_stream>
{
public:
    using value_type        = int;
    using this_type         = test_stream;
    using base_future_type  = base_future<value_type, this_type>;

public:
    int i;

    inline test_stream(int p_i)
        : i(p_i)
        { }

public:
    inline result_type poll()
    {
        return  i == 0 ? result_type::not_ready() :
                i <  0 ? result_type::done() :
                         result_type::ready(i);
    }
};

TEST(timeout_tests, poll_future_no_timeout)
{
    InSequence seq;
    StrictMock<now_mock> m;

    timing::timer tmr;
    timing::__impl::timer_base::lock l(tmr);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    test_future t(0);
    auto f = t
        .timeout<ref>(std::chrono::seconds(5));

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    auto r0 = f.poll();
    ASSERT_FALSE(r0);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    auto r1 = f.poll();
    ASSERT_FALSE(r1);

    ++t.i;

    auto r2 = f.poll();
    ASSERT_TRUE(r2);
}

TEST(timeout_tests, poll_future_timeout)
{
    InSequence seq;
    StrictMock<now_mock> m;

    timing::timer tmr;
    timing::__impl::timer_base::lock l(tmr);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    auto f = test_future(0)
        .timeout(std::chrono::seconds(5));

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    auto r0 = f.poll();
    ASSERT_FALSE(r0);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(4))));

    auto r1 = f.poll();
    ASSERT_FALSE(r1);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(5))));

    EXPECT_THROW(f.poll(), timing::timeout_exception);
}

TEST(timeout_tests, poll_stream_no_timeout)
{
    InSequence seq;
    StrictMock<now_mock> m;

    timing::timer tmr;
    timing::__impl::timer_base::lock l(tmr);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    auto t = test_stream { 0 };
    auto f = t
        .timeout<ref>(std::chrono::seconds(5));

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    auto r0 = f.poll();
    ASSERT_FALSE(r0);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    auto r1 = f.poll();
    ASSERT_FALSE(r1);

    t.i = 1;

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(4))));

    auto r2 = f.poll();
    ASSERT_TRUE(r2);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(8))));

    auto r3 = f.poll();
    ASSERT_TRUE(r3);

    t.i = 0;

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(12))));

    auto r4 = f.poll();
    ASSERT_FALSE(r4);
}

TEST(timeout_tests, poll_stream_timeout)
{
    InSequence seq;
    StrictMock<now_mock> m;

    timing::timer tmr;
    timing::__impl::timer_base::lock l(tmr);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    auto t = test_stream { 0 };
    auto f = t
        .timeout<ref>(std::chrono::seconds(5));

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(0))));

    auto r0 = f.poll();
    ASSERT_FALSE(r0);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(4))));

    auto r1 = f.poll();
    ASSERT_FALSE(r1);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(5))))
        .WillOnce(Return(time_point(std::chrono::seconds(5))));

    EXPECT_THROW(f.poll(), timing::timeout_exception);

    t.i = 1;

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(10))));

    auto r2 = f.poll();
    ASSERT_TRUE(r2);

    t.i = 0;

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(14))));

    auto r3 = f.poll();
    ASSERT_FALSE(r3);

    EXPECT_CALL(m, now())
        .WillOnce(Return(time_point(std::chrono::seconds(15))))
        .WillOnce(Return(time_point(std::chrono::seconds(15))));

    EXPECT_THROW(f.poll(), timing::timeout_exception);
}
