#include <gtest/gtest.h>

#include "../../helper/now_mock.h"
#include "../../helper/runtime_mock.h"

#include <asyncpp.h>

using namespace ::testing;
using namespace ::asyncpp;

TEST(timer_tests, lock)
{
    timing::timer t;

    EXPECT_EQ(nullptr, timing::__impl::timer_base::current());

    {
    timing::__impl::timer_base::lock l(t);

    EXPECT_EQ(&t, timing::__impl::timer_base::current());
    }

    EXPECT_EQ(nullptr, timing::__impl::timer_base::current());
}

TEST(timer_tests, resource_registration)
{
    StrictMock<now_mock> m;
    EXPECT_CALL(m, now)
        .WillRepeatedly(Return(time_point(std::chrono::seconds(0))));

    timing::timer t;
    timing::__impl::timer_base::lock l(t);

    auto f1 = std::make_unique<timing::delay>(std::chrono::seconds(10));
    auto f2 = std::make_unique<timing::delay>(std::chrono::seconds(20));
    auto f3 = std::make_unique<timing::delay>(std::chrono::seconds(30));

    EXPECT_EQ(0, t.resource_count());

    f1->poll(); // simple add

    EXPECT_EQ(1, t.resource_count());

    f2->poll(); // simple add

    EXPECT_EQ(2, t.resource_count());

    f1.reset(); // remove first

    EXPECT_EQ(1, t.resource_count());

    f3->poll(); // simple add

    EXPECT_EQ(2, t.resource_count());

    f3->poll(); // calling twice should not add

    EXPECT_EQ(2, t.resource_count());

    f3.reset(); // remove last

    EXPECT_EQ(1, t.resource_count());

    f2.reset(); // remove remaining

    EXPECT_EQ(0, t.resource_count());
}

TEST(timer_tests, idle)
{
    time_point x;
    InSequence seq;
    StrictMock<now_mock> nm;
    StrictMock<runtime_mock> rm;
    timing::timer<StrictMock<runtime_mock>&> t(rm);
    timing::__impl::timer_base::lock l(t);

    EXPECT_CALL(nm, now);
    EXPECT_CALL(rm, idle(nullptr));

    t.idle(nullptr);

    EXPECT_CALL(nm, now);
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(5)))));

    x = time_point(std::chrono::seconds(5));
    t.idle(&x);

    EXPECT_CALL(nm, now);

    auto f = std::make_unique<timing::delay>(time_point(std::chrono::seconds(10)));
    f->poll();

    EXPECT_CALL(nm, now);
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(10)))));

    t.idle(nullptr);

    EXPECT_CALL(nm, now);
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(5)))));

    x = time_point(std::chrono::seconds(5));
    t.idle(&x);

    EXPECT_CALL(nm, now);
    EXPECT_CALL(rm, idle(Pointee(time_point(std::chrono::seconds(10)))));

    x = time_point(std::chrono::seconds(15));
    t.idle(&x);
}
